from nehushtan.score.NotationTextParser import NotationTextParser
from nehushtan.score.ScoreDrawer import ScoreDrawer
from nehushtan.score.ScoreDrawerOptions import ScoreDrawerOptions

text = """
~ 1=A 4/4 | 第一百零六首 | 主的复活
5-:MP 5 6 | 6 5 5- | 5- 4 6 | 6- 5- |
>1> 祂 躺卧 在坟墓  耶 稣我 救 主
6- 7 1> | 1> 5 5- | 6- 5 #4 | 5--- |
> 静 待晨 光重曙  耶 稣我 主
1 1_.:F 1__ 3 5_. 5__ | 1>-- 1>_. 2>__ | 3> 1> 2>._ 1>__ 7_. 6__ | 5-- |
>和> 祂从 坟墓已 起 来  胜 过 仇敌大 大的 奏 凯
7_. 1>__ | 2> 2> 2>_. 1>__ 2>_. 3>__ | 1> 6 5 |
> 祂 从 黑域起 来祂 是 得胜者
 5_. 5__ | 6 6 6_. 2>__ 2>_. 1>__ | 7:RIT 1> 2> 5_. 5__ | 3>-- |
> 活 着 掌权同 祂众 圣 徒联合祂 起 来  
2>_. 1>__ | 4>-- | 3>_. 2>__ | 1> 5 3>:>RIT 2> | 1>--- ||
> 祂 起 来   阿 利 路亚主起 来
>2> 兵丁徒然看守，耶稣我救主！
> < 石头徒然封口，耶稣我主！
>3= 死亡无法锁关，耶稣我救主！
> = 祂已打断栅栏，耶稣我主！
( 1 2 ) 3. 0_ | 5_. 6__ 7~ 1> | ( 5_ 3_ 1_ )3 5<-- ||
>测> 啊哈哈哈
"""

if __name__ == '__main__':
    parser = NotationTextParser(text=text)
    parser.parse()
    parser.debug()
    font_path = "/Users/leqee/code/Verbum/fonts/Arial Unicode.ttf"

    font2 = "/Users/leqee/code/Verbum/fonts/优设好身体/YSHaoShenTi-2.ttf"

    options = ScoreDrawerOptions(font_path, 50)
    options.set_lyric_font_path(font2)
    drawer = ScoreDrawer(parser.get_parsed_score_lines(), options)
    drawer.show("test3")



import importlib
import platform
import secrets
import socket
import string
import uuid
import warnings
from typing import Union

import psutil


class CommonHelper:

    @staticmethod
    def read_target(target: Union[dict, tuple, list, None], keychain: tuple, default: any = None):
        if not keychain:
            return default

        c_target, not_find = target.copy(), object()
        for i in keychain:
            if isinstance(c_target, dict):
                c_target = c_target.get(i, not_find)
            elif isinstance(c_target, (tuple, list)):
                c_len = len(c_target)
                if isinstance(i, int) and -c_len <= i <= c_len - 1:
                    c_target = c_target[i]
                else:
                    c_target = not_find
            else:
                break
        ret = default if c_target is not_find else c_target
        return ret

    @staticmethod
    def read_dictionary(dictionary: dict, keychain: tuple, default: any = None):
        warnings.warn('DEPRECATED use read_target()', DeprecationWarning)
        return CommonHelper.read_target(dictionary, keychain, default)

    @staticmethod
    def read_array(array: tuple, keychain: tuple, default: any = None):
        warnings.warn('DEPRECATED use read_target()', DeprecationWarning)
        return CommonHelper.read_target(array, keychain, default)

    @staticmethod
    def write_dictionary(target_dict: dict, keychain: tuple, value: any):
        _target = target_dict
        for index, _key in enumerate(keychain):
            _value = _target.get(_key)
            if not isinstance(_value, dict):
                # Since 0.2.14
                # a great change: all non-dict-type entry would be cleared as an empty dict
                _value = {}
            if index == len(keychain) - 1:
                _value = value
            _target[_key] = _value
            _target = _value

        return target_dict

    @staticmethod
    def class_with_class_path(module_path: str, class_name: str = None):
        """
        Since 0.1.22
        Since 0.2.19 When the class name is the same with PY file name, `class_name` is optional.

        For a/b.py -> class b
        class_with_namespace is like 'package.sub_package.class', 'a.b'
        class_name is 'b'
        return a CLASS definition, to be used with parameters to make instance
        """
        module = importlib.import_module(module_path)
        if class_name is None:
            class_name = module_path.split('.')[-1]
        a_class = getattr(module, class_name)
        return a_class

    @staticmethod
    def class_with_module_and_name(module_base: str, sub_module_name: str):
        """
        Since 0.1.21
        Since 0.2.19 It is not so convenience to use, consider using `class_with_class_path`.

        For a/b.py -> class b
        module_base is a
        sub_module_name is b
        """
        warnings.warn('Use `class_with_class_path` instead.')

        module = __import__(module_base)
        a_class = getattr(module, sub_module_name)
        return a_class

    @staticmethod
    def generate_random_uuid_hex():
        """
        Generate a random UUID.
        Since 0.4.15
        """
        return uuid.uuid4().hex

    @staticmethod
    def generate_a_password_string(
            length=8,
            least_special_ascii_letters=-1,
            least_lower_case_letters=1,
            least_upper_case_letters=1,
            least_digits=1,
    ):
        """
        Since 0.4.21
        Parameters named as `least_*` are following one rule:
        - if it is less than 0, this kind of chars would not appear;
        - else, the total appearance count in generated password would be no less than it.
        The sum of them (if less than 0, count it as 0) should not be longer than `length`.
        """

        options = ''
        if least_lower_case_letters < 0:
            least_lower_case_letters = 0
        else:
            options += string.ascii_lowercase

        if least_upper_case_letters < 0:
            least_upper_case_letters = 0
        else:
            options += string.ascii_uppercase

        if least_digits < 0:
            least_digits = 0
        else:
            options += string.digits

        if least_special_ascii_letters < 0:
            least_special_ascii_letters = 0
        else:
            special = r"!#()*,-.:;<>@[]^_{}"
            options += special

        if least_upper_case_letters + least_lower_case_letters + least_digits + least_special_ascii_letters > length:
            raise RuntimeError("generate_secure_password error: check parameters")

        password = "".join(secrets.choice(options) for i in range(length))

        total_s = 0
        total_l = 0
        total_u = 0
        total_d = 0
        for c in password:
            if c.islower():
                total_l += 1
            elif c.isupper():
                total_u += 1
            elif c.isdigit():
                total_d += 1
            else:
                total_s += 1

        if total_d >= least_digits \
                and total_u >= least_upper_case_letters \
                and total_l >= least_lower_case_letters \
                and total_s >= least_special_ascii_letters:
            return password

        return CommonHelper.generate_a_password_string(
            length=length,
            least_special_ascii_letters=least_special_ascii_letters,
            least_lower_case_letters=least_lower_case_letters,
            least_upper_case_letters=least_upper_case_letters,
            least_digits=least_digits,
        )

    @staticmethod
    def get_python_version():
        """
        Since 0.4.25
        """
        return platform.python_version()

    @staticmethod
    def is_python_version_at_least(big_version: int, middle_version: int = 0, small_version: int = 0):
        """
        Since 0.4.25
        """
        x = platform.python_version_tuple()
        return int(x[0]) >= big_version and int(x[1]) >= middle_version and int(x[2]) >= small_version

    @staticmethod
    def get_local_ip():
        addresses = psutil.net_if_addrs()

        for interface, addr_list in addresses.items():
            for addr in addr_list:
                if addr.family == socket.AF_INET:
                    ip_address = addr.address
                    if ip_address != '127.0.0.1':
                        return ip_address

        return '127.0.0.1'



import socket
from abc import abstractmethod


class NehushtanUDPSocketClient:
    """
    Since 0.4.16
    """

    def __init__(self, host: str, port: int):
        self.__server_address = (host, port)
        self.__socket_instance: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def get_socket_instance(self):
        return self.__socket_instance

    def get_server_address(self):
        return self.__server_address

    @abstractmethod
    def handle_communication(self):
        """
        Override it to implement it with your own logic.
        use `self.__socket_instance.sendto(data, (self.__listen_host, self.__listen_port))`
        """
        pass

    def close_connection(self):
        if self.__socket_instance is not None:
            self.__socket_instance.close()
        self.__socket_instance = None



from nehushtan.httpd.NehushtanHTTPConstant import NehushtanHTTPConstant
from nehushtan.httpd.NehushtanHTTPService import NehushtanHTTPService
from nehushtan.httpd.implement.NehushtanHTTPRouteWithRegexArgs import NehushtanHTTPRouteWithRegexArgs
from nehushtan.httpd.implement.NehushtanHTTPRouteWithRegexKwargs import NehushtanHTTPRouteWithRegexKwargs
from nehushtan.httpd.implement.NehushtanHTTPRouteWithRestFul import NehushtanHTTPRouteWithRestFul
from tests.httpd.TestRequestHandler import TestRequestHandler
from tests.httpd.filter.TestFilter import TestFilter
from tests.httpd.process_chain.TestProcessChainB import TestProcessChainB

TestRequestHandler.router.register_route(
    NehushtanHTTPRouteWithRegexArgs(
        r'/read/{int}',
        ('tests.httpd.process_chain.TestProcessChainA', 'read'),
        [],
        [NehushtanHTTPConstant.METHOD_GET]
    )
).register_route(
    NehushtanHTTPRouteWithRegexArgs(
        r'/list/{str}',
        (TestProcessChainB, 'list')
    )
).register_route(
    NehushtanHTTPRouteWithRegexArgs(
        r'/step-in/{float}',
        (TestProcessChainB, 'step3'),
    )
).register_route(
    NehushtanHTTPRouteWithRegexArgs(
        r'/args/{int}/{float}/{str}/end',
        (TestProcessChainB, 'args'),
    )
).register_route(
    NehushtanHTTPRouteWithRegexKwargs(
        r'/mix/{index:int}/name/{name:str}/value/{value:float}/action/{action}',
        (TestProcessChainB, 'mix'),
    )
).register_route(
    NehushtanHTTPRouteWithRestFul(
        '/api',
        'tests.httpd.process_chain.api',
        [TestFilter]
    )
)

if __name__ == '__main__':
    NehushtanHTTPService.run_with_threading_server(TestRequestHandler, 4466)



import time
from abc import ABC

from nehushtan.mysql.MySQLTableExistence import MySQLTableExistence
from nehushtan.mysql.MySQLTableSelection import MySQLTableSelection


class MySQLViewMixin(MySQLTableExistence, ABC):
    """
    For View, actually, READONLY table.
    Since 0.1.4
    """

    @staticmethod
    def now() -> str:
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

    @staticmethod
    def today() -> str:
        return time.strftime("%Y-%m-%d", time.localtime())

    def select_in_table(self):
        return MySQLTableSelection(target=self)



import signal
from abc import abstractmethod
from typing import List


class SignalHandler:
    """
    Since 0.2.8

    If you use multi-threads, sub-processes inside shovel,
    you may need to catch the signal SIGTERM(15), which would be sent by Daemon
    to let all children die as soon as possible.
    By default, the parent process would finish when this signal comes,
    but children would not receive it and become orphaned.

    Usage
    Override this class and determine which signal(s) and a handle code block.
    The overrode class should be called in the `__init__` method of Shovel.

    See https://docs.python.org/zh-cn/3/library/signal.html?#signal.signal
    """

    @abstractmethod
    def get_target_signal_list(self) -> List[int]:
        """
        return a list of signal number, commonly use the defined constants
        signal.SIGTERM, etc.
        """
        pass

    @abstractmethod
    def handle_signal(self, signal_number, frame):
        pass

    def apply(self):
        signals = self.get_target_signal_list()
        for one_signal in signals:
            signal.signal(one_signal, self.handle_signal)

    def apply_default(self):
        signals = self.get_target_signal_list()
        for one_signal in signals:
            signal.signal(one_signal, signal.SIG_DFL)

    def apply_ignore(self):
        signals = self.get_target_signal_list()
        for one_signal in signals:
            signal.signal(one_signal, signal.SIG_IGN)



from typing import Optional, List

from psycopg2 import ProgrammingError


class PgSQLResultSet:
    """
    Since 0.4.22
    @see https://www.psycopg.org/docs/cursor.html#cursor
    """

    def __init__(self, cursor):
        self.__cursor = cursor

    def __del__(self):
        self.close()

    def get_raw_cursor(self):
        return self.__cursor

    def get_description(self):
        return self.__cursor.description

    def close(self):
        if self.__cursor:
            self.__cursor.close()

    def is_closed(self) -> bool:
        return self.__cursor.closed

    def is_scrollable(self) -> bool:
        return self.__cursor.scrollable

    def is_withhold(self) -> bool:
        return self.__cursor.withhold

    def fetch_one_row(self) -> Optional[tuple]:
        """
        Fetch the next row of a query result set,
            returning a single tuple, or None when no more data is available.
        A ProgrammingError is raised
            if the previous call to execute*() did not produce any result set or no call was issued yet.
        """
        return self.__cursor.fetchone()

    def fetch_next_batch_of_rows(self, size: int = None) -> List[tuple]:
        """
        Fetch the next set of rows of a query result,
            returning a list of tuples.
            An empty list is returned when no more rows are available.
        The number of rows to fetch per call is specified by the parameter.
            If it is not given, the cursor’s arraysize determines the number of rows to be fetched.
            The method should try to fetch as many rows as indicated by the size parameter.
            If this is not possible due to the specified number of rows not being available,
                fewer rows may be returned.
        A ProgrammingError is raised
            if the previous call to execute*() did not produce any result set or no call was issued yet.
        Note
            there are performance considerations involved with the size parameter.
            For optimal performance, it is usually best to use the arraysize attribute.
            If the size parameter is used,
                then it is best for it to retain the same value from one fetchmany() call to the next.
        """
        if size:
            return self.__cursor.fetchmany(size)
        else:
            return self.__cursor.fetchmany()

    def fetch_all_remaining_rows(self) -> List[tuple]:
        """
        Fetch all (remaining) rows of a query result, returning them as a list of tuples.
            An empty list is returned if there is no more record to fetch.
        A ProgrammingError is raised
            if the previous call to execute*() did not produce any result set or no call was issued yet.
        """
        return self.__cursor.fetchall()

    def scroll_next_n_rows(self, n: int):
        """
        Scroll the cursor in the result set to a new position according to mode.
        value is taken as offset to the current position in the result set.

        If the scroll operation would leave the result set,
            an IndexError from ProgrammingError is raised and the cursor position is not changed.

        The method can be used both for client-side cursors and server-side cursors.
            Server-side cursors can usually scroll backwards only if declared scrollable.
            Moving out-of-bound in a server-side cursor doesn’t result in an exception,
                if the backend doesn’t raise any
                (Postgres doesn’t tell us in a reliable way if we went out of bound).
        """
        try:
            self.__cursor.scroll(n, 'relative')
        except ProgrammingError as pe:
            raise IndexError from pe

    def scroll_to_target_row(self, index: int):
        """
        Scroll the cursor in the result set to a new position according to mode.
        value states an absolute target position.

        If the scroll operation would leave the result set,
            an IndexError from ProgrammingError is raised and the cursor position is not changed.

        The method can be used both for client-side cursors and server-side cursors.
            Server-side cursors can usually scroll backwards only if declared scrollable.
            Moving out-of-bound in a server-side cursor doesn’t result in an exception,
                if the backend doesn’t raise any
                (Postgres doesn’t tell us in a reliable way if we went out of bound).
        """
        try:
            self.__cursor.scroll(index, 'absolute')
        except ProgrammingError as pe:
            raise IndexError from pe

    def get_array_size_of_cursor(self) -> int:
        """
        read: This attribute specifies the number of rows to fetch at a time with fetchmany().
        It defaults to 1 meaning to fetch a single row at a time.
        """
        return self.__cursor.arraysize

    def set_array_size_of_cursor(self, x: int):
        """
        write: This attribute specifies the number of rows to fetch at a time with fetchmany().
        It defaults to 1 meaning to fetch a single row at a time.
        """
        self.__cursor.arraysize = x

    def get_iter_size_of_cursor(self) -> int:
        """
        Read: attribute specifying the number of rows to fetch from the backend
            at each network roundtrip during iteration on a named cursor.
        The default is 2000.
        """
        return self.__cursor.itersize

    def set_iter_size_of_cursor(self, x: int):
        """
        Write: attribute specifying the number of rows to fetch from the backend
            at each network roundtrip during iteration on a named cursor.
        The default is 2000.
        """
        self.__cursor.itersize = x

    def get_row_count(self) -> int:
        """
        This read-only attribute specifies the number of rows
            that the last execute*() produced (for DQL statements like SELECT)
            or affected (for DML statements like UPDATE or INSERT).
        The attribute is -1 in case no execute*() has been performed on the cursor
            or the row count of the last operation if it can’t be determined by the interface.
        Note
            The DB API 2.0 interface reserves to redefine the latter case to have the object return None
                instead of -1 in future versions of the specification.
        """
        return self.__cursor.rowcount

    def get_row_number(self) -> Optional[int]:
        """
        This read-only attribute provides the current 0-based index of the cursor in the result set
            or None if the index cannot be determined.
        The index can be seen as index of the cursor in a sequence (the result set).
            The next fetch operation will fetch the row indexed by rownumber in that sequence.
        """
        return self.__cursor.rownumber

    def get_last_inserted_row_id(self) -> Optional[int]:
        """
        This read-only attribute provides the OID of the last row inserted by the cursor.
            If the table wasn’t created with OID support or the last operation is not a single record insert,
                the attribute is set to None.
        Note PostgreSQL currently advices to not create OIDs on the tables
            and the default for CREATE TABLE is to not support them.
            The INSERT ... RETURNING syntax available from PostgreSQL 8.3 allows more flexibility.
        """
        return self.__cursor.lastrowid

    def get_query(self) -> Optional[bytes]:
        """
        Read-only attribute containing the body of the last query sent to the backend (including bound arguments)
            as bytes string.
            None if no query has been executed yet.
        """
        return self.__cursor.query

    def get_status_message(self):
        """
        Read-only attribute containing the message returned by the last command
        """
        return self.__cursor.statusmessage

    def __cast(self, oid, s):
        """
        Convert a value from the PostgreSQL string representation to a Python object.
        Use the most specific of the typecasters registered by register_type().
        """
        return self.__cursor.cast(oid, s)

    def get_tzinfo_factory(self):
        """
        The time zone factory used to handle data types such as TIMESTAMP WITH TIME ZONE.
            It should be a tzinfo object.
            Default is datetime.timezone.
        Changed in cursor version 2.9: previosly the default factory was psycopg2.tz.FixedOffsetTimezone.
        """
        return self.__cursor.tzinfo_factory

    # the following copy-series methods are not implemented now
    # copy_from(file, table, sep='\t', null='\\N', size=8192, columns=None)¶
    # copy_to(file, table, sep='\t', null='\\N', columns=None)¶
    # copy_expert(sql, file, size=8192)¶



from nehushtan.socket.wormhole.NehushtanWormhole import NehushtanWormhole

if __name__ == '__main__':
    wormhole = NehushtanWormhole(7090, '116.62.78.192', 20000, r"/Users/leqee/code/nehushtan/log")
    wormhole.run()



class NehushtanHTTPError(Exception):
    """
    Since 0.4.0
    All Exceptions raised for any Nehushtan HTTPD Service specific abnormal situations
    """

    def __init__(self, http_error_message: str, http_code: int):
        self.http_error_message = http_error_message
        self.http_code = http_code
        super().__init__(http_error_message)

    def get_http_code(self):
        return self.http_code

    def get_http_error_message(self):
        return self.http_error_message



import json
import re
import socketserver
from abc import abstractmethod
from http.server import BaseHTTPRequestHandler
from typing import Tuple, Union, Sequence, Dict
from urllib.parse import parse_qs

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.httpd.NehushtanHTTPConstant import NehushtanHTTPConstant
from nehushtan.httpd.NehushtanHTTPResponseBuffer import NehushtanHTTPResponseBuffer
from nehushtan.httpd.exceptions.NehushtanHTTPError import NehushtanHTTPError
from nehushtan.httpd.exceptions.NehushtanRequestDeniedByFilterError import NehushtanRequestDeniedByFilterError
from nehushtan.httpd.exceptions.NehushtanRequestProcessTargetError import NehushtanRequestProcessTargetError


class NehushtanHTTPRequestHandler(BaseHTTPRequestHandler):
    """
    Since 0.4.0
    """

    def __init__(self, request, client_address: Tuple[str, int], server: socketserver.BaseServer):
        # fulfilled by calling method `parse_path`
        self.parsed_path = '/'
        self.parsed_query_dict = {}
        self.parsed_cookie_dict = {}
        # fulfilled by calling method `parse_body`
        self.raw_body = b''
        self.parsed_body_data = None
        # fulfilled after do_SPAM
        self.method = ''
        # Router Matched Data
        self.matched_arguments: Sequence[str] = []
        self.matched_keyed_arguments: Dict[str, str] = {}

        # self.__process_chain_base_class = CommonHelper.class_with_class_path(
        #     'nehushtan.httpd.NehushtanHTTPRequestProcessChain'
        # )
        self.__filter_base_class = CommonHelper.class_with_class_path(
            'nehushtan.httpd.NehushtanHTTPRequestFilter'
        )
        self.__controller_base_class = CommonHelper.class_with_class_path(
            'nehushtan.httpd.NehushtanHTTPRequestController'
        )
        self.__process_chain_share_data_dict = {}

        super().__init__(request, client_address, server)

    def get_process_chain_share_data_dict(self):
        return self.__process_chain_share_data_dict

    def prepare_path_and_queries(self):
        # self.path is like path=/ab/c/d?e=f
        parts = self.path.split('?')
        self.parsed_path = parts[0]
        if len(parts) > 1:
            parsed_query_dict = parse_qs(parts[1])
            for k, v in parsed_query_dict.items():
                if k.endswith('[]'):
                    self.parsed_query_dict[k.removesuffix('[]')] = v
                elif len(v) == 1:
                    self.parsed_query_dict[k] = v[0]
                else:
                    self.parsed_query_dict[k] = v

    def prepare_cookie(self):
        # cookie
        cookie = self.headers.get('Cookie', '')
        if cookie == '':
            return
        pairs = cookie.split(";")
        for pair in pairs:
            pair = pair.lstrip()
            if pair == '':
                continue
            parts = pair.split("=", 2)
            if len(parts) != 2:
                continue
            name = parts[0].encode('latin-1').decode('utf-8')
            value = parts[1].encode('latin-1').decode('utf-8')
            self.parsed_cookie_dict[name] = value

    def prepare_body(self, body_charset=None):
        length = int(self.headers.get('content-length', 0))
        self.raw_body = self.rfile.read(length)
        # print(self.raw_body)

        content_type = self.headers.get('content-type', 'text/plain')
        # print(content_type)

        if body_charset is None:
            found_charset = re.search(r'charset=(.+);?', content_type, re.I)
            if found_charset:
                body_charset = found_charset.group(1)
            else:
                body_charset = 'utf-8'

        # print(body_charset)
        self.raw_body = self.raw_body.decode(body_charset)
        # print(self.raw_body)

        if content_type.startswith('application/x-www-form-urlencoded'):
            parsed_body_dict = parse_qs(self.raw_body)
            self.parsed_body_data = {}
            for k, v in parsed_body_dict.items():
                if k.endswith('[]'):
                    self.parsed_body_data[k.removesuffix('[]')] = v
                elif len(v) == 1:
                    self.parsed_body_data[k] = v[0]
                else:
                    self.parsed_body_data[k] = v
        elif content_type.startswith('application/json'):
            self.parsed_body_data = json.loads(self.raw_body)
        elif content_type.startswith('multipart/form-data'):
            # TODO for multipart/form-data; boundary=something
            raise NotImplementedError('BODY for FORM multipart/form-data IS NOT IMPLEMENTED')
        else:
            self.parsed_body_data = None

        # print(self.parsed_body_data, self.parsed_body_data.get('a', 'c'))

    def do_HEAD(self):
        """
        The HTTP HEAD method requests the headers
            that would be returned if the HEAD request's URL was instead requested with the HTTP GET method.
        For example, if a URL might produce a large download,
            a HEAD request could read its Content-Length header
            to check the filesize without actually downloading the file.

        Request has body	No
        Successful response has body	No
        """
        self.method = NehushtanHTTPConstant.METHOD_HEAD
        self.process_request()

    def do_GET(self):
        """
        Request has body	No
        Successful response has body	Yes
        """
        self.method = NehushtanHTTPConstant.METHOD_GET
        self.process_request()

    def do_POST(self):
        """
        @see https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/POST

        The HTTP POST method sends data to the server.
        The type of the body of the request is indicated by the Content-Type header.

        The difference between PUT and POST is that PUT is idempotent:
            calling it once or several times successively has the same effect (that is no side effect),
            where successive identical POST may have additional effects, like passing an order several times.

        A POST request is typically sent via an HTML form and results in a change on the server.
        In this case, the content type is selected by putting the adequate string
            in the enctype attribute of the <form> element
            or the formenctype attribute of the <input> or <button> elements:

        - application/x-www-form-urlencoded:
            the keys and values are encoded in key-value tuples separated by '&',
            with a '=' between the key and the value.
            Non-alphanumeric characters in both keys and values are percent encoded:
                this is the reason why this type is not suitable to use with binary data
                (use multipart/form-data instead)
        - multipart/form-data:
            each value is sent as a block of data ("body part"),
            with a user agent-defined delimiter ("boundary") separating each part.
            The keys are given in the Content-Disposition header of each part.
        - text/plain

        When the POST request is sent via a method other than an HTML form
            — like via an XMLHttpRequest —
            the body can take any type.
        As described in the HTTP 1.1 specification,
            POST is designed to allow a uniform method to cover the following functions:

        - Annotation of existing resources
        - Posting a message to a bulletin board, newsgroup, mailing list, or similar group of articles;
        - Adding a new user through a signup modal;
        - Providing a block of data, such as the result of submitting a form, to a data-handling process;
        - Extending a database through an append operation.

        Request has body	Yes
        Successful response has body	Yes
        """
        self.method = NehushtanHTTPConstant.METHOD_POST
        self.process_request()

    def do_PUT(self):
        """
        The HTTP PUT request method creates a new resource
            or replaces a representation of the target resource with the request payload.

        The difference between PUT and POST is that PUT is idempotent:
            calling it once or several times successively has the same effect (that is no side effect),
            whereas successive identical POST requests may have additional effects,
            akin to placing an order several times.

        Request has body	Yes
        Successful response has body	No
        """
        self.method = NehushtanHTTPConstant.METHOD_PUT
        self.process_request()

    def do_DELETE(self):
        """
        The HTTP DELETE request method deletes the specified resource.
        Request has body	May
        Successful response has body	May
        """
        self.method = NehushtanHTTPConstant.METHOD_DELETE
        self.process_request()

    def do_OPTIONS(self):
        """
        The HTTP OPTIONS method requests permitted communication options for a given URL or server.
        A client can specify a URL with this method, or an asterisk (*) to refer to the entire server.

        Request has body	No
        Successful response has body	Yes
        """
        self.method = NehushtanHTTPConstant.METHOD_OPTIONS
        self.process_request()

    def do_PATCH(self):
        """
        The HTTP PATCH request method applies partial modifications to a resource.

        PATCH is somewhat analogous to the "update" concept found in CRUD
            (in general, HTTP is different than CRUD, and the two should not be confused).

        A PATCH request is considered a set of instructions on how to modify a resource.
        Contrast this with PUT; which is a complete representation of a resource.

        A PATCH is not necessarily idempotent, although it can be. Contrast this with PUT;
            which is always idempotent.
        The word "idempotent" means that any number of repeated,
            identical requests will leave the resource in the same state.
        For example if an auto-incrementing counter field is an integral part of the resource,
            then a PUT will naturally overwrite it (since it overwrites everything),
            but not necessarily so for PATCH.

        PATCH (like POST) may have side-effects on other resources.

        To find out whether a server supports PATCH,
            a server can advertise its support by adding it to the list
            in the Allow or Access-Control-Allow-Methods (for CORS) response headers.

        Another (implicit) indication that PATCH is allowed,
            is the presence of the Accept-Patch header,
            which specifies the patch document formats accepted by the server.

        Request has body	Yes
        Successful response has body	Yes
        """
        self.method = NehushtanHTTPConstant.METHOD_PATCH
        self.process_request()

    def process_request(self):
        try:
            self.prepare_path_and_queries()
            self.prepare_cookie()

            filters, controller_and_method = self.seek_route_for_process_chains(self.method, self.parsed_path)

            methods_contains_body = (
                NehushtanHTTPConstant.METHOD_POST,
                NehushtanHTTPConstant.METHOD_PUT,
                NehushtanHTTPConstant.METHOD_DELETE,
                NehushtanHTTPConstant.METHOD_PATCH,
            )
            if methods_contains_body.__contains__(self.method):
                self.prepare_body()

            filter_data_dict = {}
            for filter_item in filters:
                c = filter_item
                if type(c) is str:
                    c = CommonHelper.class_with_class_path(c)
                if not issubclass(c, self.__filter_base_class):
                    raise NehushtanRequestProcessTargetError(
                        http_error_message=f'Filter Class Error: {filter_item}',
                        http_code=500
                    )
                filter_instance = c(self, filter_data_dict)
                if not filter_instance.should_accept_request():
                    raise NehushtanRequestDeniedByFilterError(
                        filter_item,
                        filter_instance.message_for_denial,
                        filter_instance.http_code_for_denial
                    )

            controller_item = controller_and_method[0]
            method_item = controller_and_method[1]

            c = controller_item
            if type(c) is str:
                c = CommonHelper.class_with_class_path(c)
            if not issubclass(c, self.__controller_base_class):
                raise NehushtanRequestProcessTargetError(
                    http_error_message=f'Controller Class Error for {controller_item}',
                    http_code=500
                )
            controller_instance = c(self)
            controller_instance_target_method = getattr(controller_instance, method_item)

            args = self.matched_arguments
            kwargs = self.matched_keyed_arguments
            controller_instance_target_method(*args, **kwargs)

        except NehushtanHTTPError as http_error:
            self.send_error(http_error.get_http_code(), http_error.http_error_message)
            self.wfile.write(http_error.get_http_error_message().encode())

    @abstractmethod
    def seek_route_for_process_chains(self, method: str, path: str) \
            -> Tuple[Sequence[Union[type, str]], Tuple[Union[type, str], str]]:
        """
        You may need to update `self.matched_arguments` or `self.matched_keyed_arguments` here.
        It would return the filter list and the controller-method pair.
        """
        pass

    def send_response_with_buffer(self, buffer: NehushtanHTTPResponseBuffer):
        if buffer.http_code < 200 or buffer.http_code >= 400:
            self.send_error(buffer.http_code, buffer.http_code_message)
            return self

        self.send_response(buffer.http_code, buffer.http_code_message)

        text = buffer.get_body_as_string()

        if buffer.encoding is not None:
            s = text.encode(buffer.encoding)
        else:
            s = text.encode()

        buffer.set_header('Content-Length', '%i' % len(s))

        for k, v in buffer.headers.items():
            self.send_header(k, v)
        self.end_headers()

        self.wfile.write(s)



from nehushtan.wave.Note import Note
from nehushtan.wave.SimpleComposer import SimpleComposer


def write_part_v1(composer: SimpleComposer):
    composer.add_note_to_line(Note(name="E"))
    composer.add_note_to_line(Note(name="E"))
    composer.add_note_to_line(Note(name="F"))
    composer.add_note_to_line(Note(name="G"))
    composer.add_note_to_line(Note(name="G"))
    composer.add_note_to_line(Note(name="F"))
    composer.add_note_to_line(Note(name="E"))
    composer.add_note_to_line(Note(name="D"))
    composer.add_note_to_line(Note(name="C"))
    composer.add_note_to_line(Note(name="C"))
    composer.add_note_to_line(Note(name="D"))
    composer.add_note_to_line(Note(name="E"))
    composer.add_note_to_line(Note(name="E", beats=1.5))
    composer.add_note_to_line(Note(name="D", beats=0.5))
    composer.add_note_to_line(Note(name="D", beats=2))

    composer.add_note_to_line(Note(name="G", group=4), 1)
    composer.add_note_to_line(Note(name="C", group=5), 1)
    composer.add_note_to_line(Note(name="A", group=4), 1)
    composer.add_note_to_line(Note(name="C", group=5), 1)
    composer.add_note_to_line(Note(name="G", group=5), 1)
    composer.add_note_to_line(Note(name="F", group=5), 1)
    composer.add_note_to_line(Note(name="E", group=5), 1)
    composer.add_note_to_line(Note(name="D", group=5), 1)
    composer.add_note_to_line(Note(name="C", group=5), 1)
    composer.add_note_to_line(Note(name="C", group=5), 1)
    composer.add_note_to_line(Note(name="D", group=5), 1)
    composer.add_note_to_line(Note(name="E", group=5), 1)
    composer.add_note_to_line(Note(name="E", group=5, beats=1.5), 1)
    composer.add_note_to_line(Note(name="D", group=5, beats=0.5), 1)
    composer.add_note_to_line(Note(name="D", group=5, beats=2), 1)


def write_part_v2(composer: SimpleComposer):
    composer.add_notes_to_lines([
        Note(name="E"),
        Note(name="C", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="E"),
        Note(name="G", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="F"),
        Note(name="A", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="G"),
        Note(name="B", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="G"),
        Note(name="B", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="F"),
        Note(name="A", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="E"),
        Note(name="G", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="D"),
        Note(name="F", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="C"),
        Note(name="C", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="C"),
        Note(name="E", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="D"),
        Note(name="G", group=3),
    ])
    composer.add_notes_to_lines([
        Note(name="E"),
        Note(name="E", group=3),
    ])
    composer.add_note_to_line(Note(name="E", beats=1.5), 0)
    composer.add_note_to_line(Note(name="C", group=3, beats=0.5), 1)
    composer.add_note_to_line(Note(name="E", group=3, beats=0.5), 1)
    composer.add_note_to_line(Note(name="G", group=3, beats=0.5), 1)

    composer.add_notes_to_lines([
        Note(name="D", beats=0.5),
        Note(name="D", group=3, beats=0.5),
    ])

    composer.add_note_to_line(Note(name="D", beats=2), 0)
    composer.add_note_to_line(Note(name="G", group=3, beats=0.5), 1)
    composer.add_note_to_line(Note(name="B", group=3, beats=1.5), 1)


if __name__ == '__main__':
    composer = SimpleComposer(r"/Users/leqee/code/nehushtan/debug/wave/test-1.wav")
    write_part_v2(composer)
    composer.close()



import asyncio

from nehushtan.logger.NehushtanLogger import NehushtanLogger

logger = NehushtanLogger()


async def async_func_1(task_id: int):
    logger.info(f"async_func_1 start: {task_id}")
    await asyncio.sleep(1)
    logger.info(f"async_func_1 end: {task_id}")


async def async_func_0():
    task_dict = {}
    for i in range(10):
        # use create_task
        x = asyncio.create_task(async_func_1(i))
        task_dict[i] = x

    logger.info("TASKS CREATED")
    for i in range(10):
        await task_dict[i]

    logger.info("TASKS AWAITED")


if __name__ == '__main__':
    logger.info("main start")
    # they would run in order
    # for i in range(10):
    #     asyncio.run(async_func_1(i))

    asyncio.run(async_func_0())

    logger.info("main end")



from typing import Optional

from nehushtan.socket.NehushtanUDPSocketClient import NehushtanUDPSocketClient
from tests.socket.badmq.MQCommand import MQCommand


class BadMQClient(NehushtanUDPSocketClient):

    def __init__(self, host: str, port: int):
        super().__init__(host, port)
        self.__command: Optional[MQCommand] = None

    def handle_communication(self):
        sent = self.get_socket_instance().sendto(self.__command.package_as_bytes(), self.get_server_address())
        print(f'[DEBUG] sent to {self.get_server_address()}, {sent} bytes: {self.__command.package_as_bytes()}')

        data = self.get_socket_instance().recv(1024)
        self.__command = MQCommand.parse(data)
        if not self.__command:
            print(f'[WARN] RESULT CANNOT BE PARSED: {data}')
        else:
            print(f'[INFO] RESULT: {self.__command.package_as_bytes()}')

    def dequeue(self):
        self.__command = MQCommand(MQCommand.COMMAND_DEQUEUE, 'QueueA')
        self.handle_communication()
        if self.__command:
            return self.__command.value
        else:
            raise IOError('pop failed')

    def enqueue(self, xx: str):
        self.__command = MQCommand(MQCommand.COMMAND_ENQUEUE, 'QueueA', xx)
        self.handle_communication()
        if self.__command:
            return self.__command.value
        else:
            raise IOError('pop failed')


if __name__ == '__main__':
    x = BadMQClient('127.0.0.1', 6767)
    print('enqueue 1:', x.enqueue('1'))
    print('enqueue 2:', x.enqueue('2'))
    print('enqueue 3:', x.enqueue('3'))
    print('dequeue (1)', x.dequeue())
    print('dequeue (2)', x.dequeue())
    print('dequeue (3)', x.dequeue())
    print('dequeue (?)', x.dequeue())
